import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Point2D;
import java.util.*;

public abstract class Entity
{
	protected Sprite sprite;
	protected Point2D.Double pos;
	protected int inner;
	protected int outer;
	protected int direction;
	
	public Entity (Point2D.Double start, int inner)
	{
		this(start, inner, inner);
	}	
	
	public Entity (Point2D.Double start, int inner, int outer)
	{
		this(start, inner, outer, 0);
	}	
	
	public Entity (Point2D.Double start, int inner, int outer, int direction)
	{
		this.pos = start;
		this.inner = inner;
		this.outer = outer;
		this.direction = direction;
	}
	
	public void setSprite (Sprite sprite)
	{
		this.sprite = sprite;
	}
	
	public Sprite getSprite ()
	{
		return sprite;
	}
	
	public void setPosition (Point2D.Double p)
	{
		pos = p;
	}
	
	public Point2D.Double getPosition ()
	{
		return pos;
	}
	
	public void setDirection (int d)
	{
		direction = d;
	}
	
	public int getDirection ()
	{
		return direction;
	}
	
	public int leftEdge ()
	{
		return (int) pos.x - outer;
	}
	
	public int topEdge ()
	{
		return (int) pos.y - outer;
	}
	
	public int inner ()
	{
		return inner;
	}
	
	public int outer ()
	{
		return outer;
	}
	
	public boolean valid (double x, double y)
	{
		Point2D.Double p = new Point2D.Double(x, y);
		
		return valid(p);
	}
	
	public boolean valid (Point2D.Double p)
	{
		Point2D.Double[] points = corners(p);
		
		for (int i = 0; i < points.length; i++)
		{
			Tile tile = GameArea.getLevel().getTileAt(points[i]);
			
			if (this instanceof Projectile)
			{
				if (tile.type == "wall")
				{
					return false;
				}
			}
			else
			{
				if (tile.type != "floor")
				{
					return false;
				}
			}
		}
		
		for (Entity entity: GameArea.getLevel().entities)
		{
			if (wouldTouch(p, entity) && this != entity)
			{
				if (collision(entity))
				{
					return false;
				}
			}
		}
		
		return true;
	}
	
	public boolean collision (Entity entity)
	{
		return (entity.solid());
	}
	
	public Point2D.Double[] corners (Point2D.Double p)
	{
		Point2D.Double[] points = new Point2D.Double[4];
		
		points[0] = new Point2D.Double(p.x - inner, p.y - inner);
		points[1] = new Point2D.Double(p.x - inner, p.y + inner);
		points[2] = new Point2D.Double(p.x + inner, p.y - inner);
		points[3] = new Point2D.Double(p.x + inner, p.y + inner);
		
		return points;
	}
	
	public boolean move (double distance)
	{
		return move(distance, 0);
	}
	
	public boolean move (double distance, int offsetDirection)
	{
		double dx = Math.cos(radians(direction - offsetDirection)) * distance;
		double dy = Math.sin(radians(direction - offsetDirection)) * distance;
		
		if (this instanceof Player)
		{
			boolean moveX = valid(pos.x + dx, pos.y);
			boolean moveY = valid(pos.x, pos.y + dy);
			boolean moveBoth = valid(pos.x + dx, pos.y + dy);
			
			if (! moveBoth)
			{
				if (moveX && moveY)
				{
					if (Math.abs(dx) > Math.abs(dy))
					{
						moveY = false;
					}
					else
					{
						moveX = false;
					}
				}
				
				if (! moveX)
				{
					dx = 0;
				}
				
				if (! moveY)
				{
					dy = 0;
				}
			}
			
			pos.x += dx;
			pos.y += dy;
			
			// One of the directions is valid
			return (dx != 0 || dy != 0);
		}
		else
		{
			Point2D.Double p = new Point2D.Double(pos.x + dx, pos.y + dy);
			
			if (valid(p))
			{
				pos = p;
				
				return true;
			}
			else
			{
				return false;
			}
		}
	}
	
	public void turnLeft ()
	{
		direction = (direction + 47) % 48;
	}
	
	public void turnRight ()
	{
		direction = (direction + 1) % 48;
	}
	
	public boolean solid ()
	{
		return false;
	}
	
	public void draw (Graphics g, int x, int y)
	{
		if (! (g instanceof Graphics2D))
		{
			System.out.println("Oh dear.");
			return;
		}
		
		Graphics2D g2 = (Graphics2D) g;
		
		int relativeX = leftEdge() - x;
		int relativeY = topEdge() - y;
		
		g2.rotate(radians(direction), relativeX + outer, relativeY + outer);
		
		sprite.draw(g2, relativeX, relativeY);
		
		g2.rotate(- radians(direction), relativeX + outer, relativeY + outer);
	}
	
	public static double distance (Entity a, Entity b)
	{
		return distance(a.getPosition(), b.getPosition());
	}
	
	public static double distance (Entity a, Point2D.Double b)
	{
		return distance(a.getPosition(), b);
	}
	
	public static double distance (Point2D.Double a, Entity b)
	{
		return distance(a, b.getPosition());
	}
	
	public static double distance (Point2D.Double a, Point2D.Double b)
	{
		return a.distance(b);
	}
	
	public int directionTo (Entity entity)
	{
		return directionTo(entity.getPosition());
	}
	
	public int directionTo (Point2D.Double there)
	{
		double xDist = there.x - pos.x;
		double yDist = there.y - pos.y;
		
		return (int) (Math.atan2(yDist, xDist) / Math.PI * 24);
	}
	
	public boolean touching (Entity other)
	{
		return wouldTouch(pos, other);
	}
	
	public boolean wouldTouch (Point2D.Double p, Entity other)
	{
		return (distance(p, other) < inner + other.inner());
	}
	
	public abstract void cycle ();
	
	public static double radians (int direction)
	{
		return (Math.PI * direction) / 24;
	}
}
